@@ -25,6 +25,10 @@
   @NOTATION@
  */
 
+#undef TIME_INIT
+
+#define GNOME_ACCESSIBILITY_ENV "GNOME_ACCESSIBILITY"
+#define GNOME_ACCESSIBILITY_KEY "/desktop/gnome/interface/accessibility"
 
 /* This module takes care of handling application and library
    initialization and command line parsing */
@@ -32,12 +36,14 @@
 #include <config.h>
 #include "gnome-macros.h"
 
-#include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <gmodule.h>
+#include <gconf/gconf.h>
+#include <gconf/gconf-value.h>
+#include <gconf/gconf-client.h>
 
 #include "gnome-i18nP.h"
 
@@ -47,11 +53,6 @@
 #include "gnome-i18n.h"
 #include "gnome-url.h"
 
-#include <gobject/gboxed.h>
-#include <gobject/gvaluearray.h>
-#include <gobject/gvaluetypes.h>
-#include <gobject/gparamspecs.h>
-
 struct _GnomeProgramPrivate {
     enum {
 	APP_UNINIT=0,
@@ -61,8 +62,6 @@ struct _GnomeProgramPrivate {
     } state;
 
     /* Construction properties */
-    GnomeModuleInfo *prop_module_info;
-    gchar *prop_module_list;
     int prop_popt_flags;
     struct poptOptions *prop_popt_table;
     gchar *prop_human_readable_name;
@@ -89,20 +88,13 @@ struct _GnomeProgramPrivate {
     /* valid-while: state == APP_PREINIT_DONE */
     poptContext arg_context;
 
-    /* valid-while: state == APP_PREINIT_DONE */
-    struct poptOptions *app_options;
-
-    /* valid-while: state == APP_PREINIT_DONE */
-    int app_popt_flags;
-
     /* valid-while: state == APP_PREINIT_DONE */
     GArray *top_options_table;
+    GSList *accessibility_modules;
 };
 
 enum {
     PROP_0,
-    PROP_MODULE_INFO,
-    PROP_MODULES,
     PROP_APP_ID,
     PROP_APP_VERSION,
     PROP_HUMAN_READABLE_NAME,
@@ -124,8 +116,6 @@ enum {
     PROP_LAST
 };
 
-static void gnome_program_class_init    (GnomeProgramClass *klass);
-static void gnome_program_instance_init (GnomeProgram      *program);
 static void gnome_program_finalize      (GObject           *object);
 
 static GQuark quark_get_prop = 0;
@@ -142,7 +132,7 @@ static guint last_property_id = PROP_LAST;
 #define	PREALLOC_MODINFOS (8)
 
 GNOME_CLASS_BOILERPLATE (GnomeProgram, gnome_program,
-			 GObject, g_object, G_TYPE_OBJECT)
+			 GObject, G_TYPE_OBJECT)
 
 static void
 gnome_program_set_property (GObject *object, guint param_id,
@@ -156,13 +146,6 @@ gnome_program_set_property (GObject *object, guint param_id,
     program = GNOME_PROGRAM (object);
 
     switch (param_id) {
-    case PROP_MODULE_INFO:
-	program->_priv->prop_module_info = g_value_dup_boxed ((GValue *) value);
-	break;
-    case PROP_MODULES:
-	g_free (program->_priv->prop_module_list);
-	program->_priv->prop_module_list = g_value_dup_string (value);
-	break;
     case PROP_POPT_TABLE:
 	program->_priv->prop_popt_table = g_value_peek_pointer (value);
 	break;
@@ -360,7 +343,7 @@ gnome_program_module_addtolist (GnomeModuleInfo **new_list,
     GnomeModuleInfo *new_item;
     int i;
 
-    g_assert (new_item >= 0);
+    g_assert (new_item_idx >= 0);
 
     new_item = g_ptr_array_index (program_modules, new_item_idx);
     if(!new_item)
@@ -436,37 +419,30 @@ gnome_program_class_init (GnomeProgramClass *class)
     object_class->set_property = gnome_program_set_property;
     object_class->get_property = gnome_program_get_property;
 
-    g_object_class_install_property
-	(object_class,
-	 PROP_MODULE_INFO,
-	 g_param_spec_boxed (GNOME_PARAM_MODULE_INFO, NULL, NULL,
-			     GNOME_TYPE_MODULE_INFO,
-			     (G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY)));
-
-    g_object_class_install_property
-	(object_class,
-	 PROP_MODULES,
-	 g_param_spec_string (GNOME_PARAM_MODULES, NULL, NULL,
-			      NULL,
-			      (G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY)));
-
     g_object_class_install_property
 	(object_class,
 	 PROP_POPT_TABLE,
-	 g_param_spec_pointer (GNOME_PARAM_POPT_TABLE, NULL, NULL,
+	 g_param_spec_pointer (GNOME_PARAM_POPT_TABLE,
+			       _("Popt Table"), 
+			       _("The table of options for popt"),
 			       (G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY)));
 
     g_object_class_install_property
 	(object_class,
 	 PROP_POPT_FLAGS,
-	 g_param_spec_int (GNOME_PARAM_POPT_FLAGS, NULL, NULL,
+	 g_param_spec_int (GNOME_PARAM_POPT_FLAGS,
+			   _("Popt Flags"), 
+			   _("The flags to use for popt"),
 			   G_MININT, G_MAXINT, 0,
 			   (G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY)));
 
     g_object_class_install_property
 	(object_class,
 	 PROP_POPT_CONTEXT,
-	 g_param_spec_pointer (GNOME_PARAM_POPT_CONTEXT, NULL, NULL,
+	 g_param_spec_pointer (GNOME_PARAM_POPT_CONTEXT,
+			      _("Popt Context"), 
+			      _("The popt context pointer that GnomeProgram "
+				"is using"),
 			       (G_PARAM_READABLE)));
 
     g_object_class_install_property
@@ -492,7 +468,9 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_APP_ID,
-	 g_param_spec_string (GNOME_PARAM_APP_ID, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_APP_ID,
+			      _("App ID"), 
+			      _("ID string to use for this application"),
 			      NULL, G_PARAM_READABLE));
 
     g_object_class_install_property
@@ -506,7 +484,9 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_GNOME_PREFIX,
-	 g_param_spec_string (GNOME_PARAM_GNOME_PREFIX, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_GNOME_PREFIX,
+			      _("GNOME Prefix"), 
+			      _("Prefix where GNOME was installed"),
 			      LIBGNOME_PREFIX,
 			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
 			       G_PARAM_CONSTRUCT_ONLY)));
@@ -514,7 +494,9 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_GNOME_LIBDIR,
-	 g_param_spec_string (GNOME_PARAM_GNOME_LIBDIR, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_GNOME_LIBDIR,
+			      _("GNOME Libdir"), 
+			      _("Library prefix where GNOME was installed"),
 			      LIBGNOME_LIBDIR,
 			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
 			       G_PARAM_CONSTRUCT_ONLY)));
@@ -522,7 +504,9 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_GNOME_DATADIR,
-	 g_param_spec_string (GNOME_PARAM_GNOME_DATADIR, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_GNOME_DATADIR,
+			      _("GNOME Datadir"), 
+			      _("Data prefix where GNOME was installed"),
 			      LIBGNOME_DATADIR,
 			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
 			       G_PARAM_CONSTRUCT_ONLY)));
@@ -530,7 +514,10 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_GNOME_SYSCONFDIR,
-	 g_param_spec_string (GNOME_PARAM_GNOME_SYSCONFDIR, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_GNOME_SYSCONFDIR,
+			      _("GNOME Sysconfdir"), 
+			      _("Configuration prefix where GNOME "
+				"was installed"),
 			      LIBGNOME_SYSCONFDIR,
 			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
 			       G_PARAM_CONSTRUCT_ONLY)));
@@ -538,39 +525,48 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_APP_PREFIX,
-	 g_param_spec_string (GNOME_PARAM_APP_PREFIX, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_APP_PREFIX,
+			      _("GNOME App Prefix"), 
+			      _("Prefix where this application was installed"),
 			      NULL,
-			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
-			       G_PARAM_CONSTRUCT_ONLY)));
+			      (G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
     g_object_class_install_property
 	(object_class,
 	 PROP_APP_LIBDIR,
-	 g_param_spec_string (GNOME_PARAM_APP_LIBDIR, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_APP_LIBDIR,
+			      _("GNOME App Libdir"), 
+			      _("Library prefix where this application "
+				"was installed"),
 			      NULL,
-			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
-			       G_PARAM_CONSTRUCT_ONLY)));
+			      (G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
     g_object_class_install_property
 	(object_class,
 	 PROP_APP_DATADIR,
-	 g_param_spec_string (GNOME_PARAM_APP_DATADIR, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_APP_DATADIR,
+			      _("GNOME App Datadir"), 
+			      _("Data prefix where this application "
+				"was installed"),
 			      NULL,
-			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
-			       G_PARAM_CONSTRUCT_ONLY)));
+			      (G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
     g_object_class_install_property
 	(object_class,
 	 PROP_APP_SYSCONFDIR,
-	 g_param_spec_string (GNOME_PARAM_APP_SYSCONFDIR, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_APP_SYSCONFDIR,
+			      _("GNOME App Sysconfdir"), 
+			      _("Configuration prefix where this application "
+				"was installed"),
 			      NULL,
-			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
-			       G_PARAM_CONSTRUCT_ONLY)));
+			      (G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
     g_object_class_install_property
 	(object_class,
 	 PROP_CREATE_DIRECTORIES,
-	 g_param_spec_boolean (GNOME_PARAM_CREATE_DIRECTORIES, NULL, NULL,
+	 g_param_spec_boolean (GNOME_PARAM_CREATE_DIRECTORIES,
+			      _("Create Directories"), 
+			      _("Create standard GNOME directories on startup"),
 			       TRUE,
 			       (G_PARAM_READABLE | G_PARAM_WRITABLE |
 				G_PARAM_CONSTRUCT_ONLY)));
@@ -578,7 +574,9 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_ENABLE_SOUND,
-	 g_param_spec_boolean (GNOME_PARAM_ENABLE_SOUND, NULL, NULL,
+	 g_param_spec_boolean (GNOME_PARAM_ENABLE_SOUND,
+			      _("Enable Sound"), 
+			      _("Enable sound on startup"),
 			       TRUE,
 			       (G_PARAM_READABLE | G_PARAM_WRITABLE |
 				G_PARAM_CONSTRUCT_ONLY)));
@@ -586,7 +584,9 @@ gnome_program_class_init (GnomeProgramClass *class)
     g_object_class_install_property
 	(object_class,
 	 PROP_ESPEAKER,
-	 g_param_spec_string (GNOME_PARAM_ESPEAKER, NULL, NULL,
+	 g_param_spec_string (GNOME_PARAM_ESPEAKER,
+			      _("Espeaker"), 
+			      _("How to connect to esd"),
 			      NULL,
 			      (G_PARAM_READABLE | G_PARAM_WRITABLE |
 			       G_PARAM_CONSTRUCT_ONLY)));
@@ -607,13 +607,17 @@ gnome_program_instance_init (GnomeProgram *program)
 	GnomeModuleInfo *a_module = g_ptr_array_index (program_modules, i);
 
 	if (a_module && a_module->instance_init) {
+#ifdef TIME_INIT
 	    GTimer *timer = g_timer_new ();
 	    g_timer_start (timer);
 	    g_print ("Running class_init for: %s ...", a_module->name); 
+#endif
 	    a_module->instance_init (program, a_module);
+#ifdef TIME_INIT
 	    g_timer_stop (timer);
 	    g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 	    g_timer_destroy (timer);
+#endif
 	}
     }
 }
@@ -621,26 +625,59 @@ gnome_program_instance_init (GnomeProgram *program)
 static void
 gnome_program_finalize (GObject* object)
 {
-    GnomeProgram *program = GNOME_PROGRAM (object);
+	GnomeProgram *self = GNOME_PROGRAM (object);
 
-    /* FIXME! free everything, but perhaps this is
-     * useless, we never get finalized */
+	/* no free */
+	self->_priv->prop_popt_table = NULL;
 
-    g_free (program->_priv);
-    program->_priv = NULL;
-  
-    if (G_OBJECT_CLASS (parent_class)->finalize)
-	(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+	g_free (self->_priv->prop_human_readable_name);
+	self->_priv->prop_human_readable_name = NULL;
+	g_free (self->_priv->prop_gnome_prefix);
+	self->_priv->prop_gnome_prefix = NULL;
+	g_free (self->_priv->prop_gnome_libdir);
+	self->_priv->prop_gnome_libdir = NULL;
+	g_free (self->_priv->prop_gnome_sysconfdir);
+	self->_priv->prop_gnome_sysconfdir = NULL;
+	g_free (self->_priv->prop_gnome_datadir);
+	self->_priv->prop_gnome_datadir = NULL;
+	g_free (self->_priv->prop_app_prefix);
+	self->_priv->prop_app_prefix = NULL;
+	g_free (self->_priv->prop_app_libdir);
+	self->_priv->prop_app_libdir = NULL;
+	g_free (self->_priv->prop_app_sysconfdir);
+	self->_priv->prop_app_sysconfdir = NULL;
+	g_free (self->_priv->prop_app_datadir);
+	self->_priv->prop_app_datadir = NULL;
+	g_free (self->_priv->prop_espeaker);
+	self->_priv->prop_espeaker = NULL;
+
+	g_strfreev (self->_priv->gnome_path);
+	self->_priv->gnome_path = NULL;
+
+	g_free (self->_priv->app_id);
+	self->_priv->app_id = NULL;
+	g_free (self->_priv->app_version);
+	self->_priv->app_version = NULL;
+
+	g_strfreev (self->_priv->argv);
+	self->_priv->argv = NULL;
+
+	if (self->_priv->arg_context != NULL) {
+		poptFreeContext (self->_priv->arg_context);
+		g_dataset_destroy (self->_priv->arg_context);
+	}
+	self->_priv->arg_context = NULL;
+
+	if (self->_priv->top_options_table != NULL)
+		g_array_free (self->_priv->top_options_table, TRUE);
+	self->_priv->top_options_table = NULL;
+
+	g_free (self->_priv);
+	self->_priv = NULL;
+
+	GNOME_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
 }
 
-/* UNUSED
-static gpointer
-gnome_module_info_init (void)
-{
-    return g_new0 (GnomeModuleInfo, 1);
-}
-*/
-
 static gpointer
 gnome_module_info_copy (gpointer boxed)
 {
@@ -669,10 +706,12 @@ gnome_module_info_get_type (void)
 /**
  * gnome_program_get:
  *
- * Returns: An object that stores information on the GNOME application's state.
- * If the object does not exist, %NULL is returned. Other functions assume that
- * this will always return a #GnomeProgram object which (if not %NULL) has
- * already been initialized.
+ * Retrieves an object that stored information about the application's state.
+ * Other functions assume this will always return a #GnomeProgram object which
+ * (if not %NULL) has already been initialized.
+ *
+ * Returns: The application's #GnomeProgram instance, or %NULL if it does not
+ * exist.
  */
 
 GnomeProgram *
@@ -744,9 +783,27 @@ gnome_program_get_human_readable_name (GnomeProgram *program)
     g_return_val_if_fail (GNOME_IS_PROGRAM (program), NULL);
     g_return_val_if_fail (program->_priv->state >= APP_PREINIT_DONE, NULL);
 
+    if (program->_priv->prop_human_readable_name == NULL)
+      return g_get_prgname ();
+    
     return program->_priv->prop_human_readable_name;
 }
 
+/**
+ * gnome_program_install_property:
+ * @pclass: A #GnomeProgramClass.
+ * @get_fn: A function to get property values.
+ * @set_fn: A function to set property values.
+ * @pspec: A collection of properties.
+ *
+ * Install a collection of available properties, their default values and the
+ * functions to set and retrieve these properties.
+ *
+ * Normal applications will never need to call this function, it is mostly for
+ * use by other platform library authors.
+ *
+ * Returns: The number of properties installed.
+ */
 guint
 gnome_program_install_property (GnomeProgramClass *pclass,
 				GObjectGetPropertyFunc get_fn,
@@ -777,17 +834,20 @@ gnome_program_install_property (GnomeProgramClass *pclass,
  * @ret_locations: If this is not %NULL, a list of all the possible locations
  *                 of the file will be returned.
  *
- * This function finds the full path to a file located in the specified
+ * This function finds a full path to the @file_name located in the specified
  * "domain". A domain is a name for a collection of related files.
  * For example, common domains are "libdir", "pixmap", and "config".
  *
- * If @ret_locations is %NULL, there is no other return value.
+ * If @ret_locations is %NULL, only one pathname is returned. Otherwise,
+ * alternative paths are returned in @ret_locations.
  *
- * The GNOME_FILE_DOMAIN_APP_* domains are ones for your own application.
- * However you MUST set the correct attributes for #GnomeProgram for the APP
- * specific prefixes (during the initialization part of the application).
+ * User applications should store files in the GNOME_FILE_DOMAIN_APP_*
+ * domains. However you MUST set the correct attributes for #GnomeProgram for
+ * the APP specific prefixes (during the initialization part of the
+ * application).
  *
- * The @ret_locations list and its contents should be freed by the caller.
+ * The @ret_locations list and its contents should be freed by the caller, as
+ * should the returned string.
  *
  * Returns: The full path to the file (if it exists or only_if_exists is
  *          %FALSE) or %NULL.
@@ -820,6 +880,13 @@ gnome_program_locate_file (GnomeProgram *program, GnomeFileDomain domain,
 	} \
 }
 
+    /* Potentially add an absolute path */
+    if (g_path_is_absolute (file_name))
+      {
+        if (!only_if_exists || g_file_test (file_name, G_FILE_TEST_EXISTS))
+          ADD_FILENAME (file_name);
+      }
+    
     switch (domain) {
     case GNOME_FILE_DOMAIN_LIBDIR:
 	prefix_rel = "/lib";
@@ -882,21 +949,21 @@ gnome_program_locate_file (GnomeProgram *program, GnomeFileDomain domain,
 	search_path = FALSE;
 	break;
     case GNOME_FILE_DOMAIN_APP_HELP:
-	len = strlen ("/share//help") + 
+	len = strlen ("/share/gnome/help/") + 
 		strlen (program->_priv->app_id) + 1;
 	prefix_rel = g_alloca (len);
 	if (prefix_rel == NULL /* bad things */)
 		return NULL;
-	g_snprintf (prefix_rel, len, "/share/%s/help", program->_priv->app_id);
+	g_snprintf (prefix_rel, len, "/share/gnome/help/%s", program->_priv->app_id);
 
 	attr_name = GNOME_PARAM_APP_DATADIR;
 
-	len = strlen ("//help") + 
+	len = strlen ("/gnome/help/") + 
 		strlen (program->_priv->app_id) + 1;
 	attr_rel = g_alloca (len);
 	if (attr_rel == NULL /* bad things */)
 		return NULL;
-	g_snprintf (attr_rel, len, "/%s/help", program->_priv->app_id);
+	g_snprintf (attr_rel, len, "/gnome/help/%s", program->_priv->app_id);
 
 	search_path = FALSE;
 	break;
@@ -924,6 +991,7 @@ gnome_program_locate_file (GnomeProgram *program, GnomeFileDomain domain,
 	    g_snprintf (fnbuf, sizeof (fnbuf), "%s%s/%s",
 			dir, attr_rel, file_name);
 
+	    g_free (dir);
 	    if (!only_if_exists || g_file_test (fnbuf, G_FILE_TEST_EXISTS))
 		ADD_FILENAME (fnbuf);
 	}
@@ -979,8 +1047,8 @@ static int rpmvercmp(const char * a, const char * b) {
 
     /* loop through each version segment of str1 and str2 and compare them */
     while (*one && *two) {
-	while (*one && !isalnum(*one)) one++;
-	while (*two && !isalnum(*two)) two++;
+	while (*one && !g_ascii_isalnum(*one)) one++;
+	while (*two && !g_ascii_isalnum(*two)) two++;
 
 	str1 = one;
 	str2 = two;
@@ -988,13 +1056,13 @@ static int rpmvercmp(const char * a, const char * b) {
 	/* grab first completely alpha or completely numeric segment */
 	/* leave one and two pointing to the start of the alpha or numeric */
 	/* segment and walk str1 and str2 to end of segment */
-	if (isdigit(*str1)) {
-	    while (*str1 && isdigit(*str1)) str1++;
-	    while (*str2 && isdigit(*str2)) str2++;
+	if (g_ascii_isdigit(*str1)) {
+	    while (*str1 && g_ascii_isdigit(*str1)) str1++;
+	    while (*str2 && g_ascii_isdigit(*str2)) str2++;
 	    isnum = 1;
 	} else {
-	    while (*str1 && isalpha(*str1)) str1++;
-	    while (*str2 && isalpha(*str2)) str2++;
+	    while (*str1 && g_ascii_isalpha(*str1)) str1++;
+	    while (*str2 && g_ascii_isalpha(*str2)) str2++;
 	    isnum = 0;
 	}
 		
@@ -1162,6 +1230,19 @@ gnome_program_module_register (const GnomeModuleInfo *module_info)
     }
 }
 
+static void
+set_context_data (poptContext con,
+		  enum poptCallbackReason reason,
+		  const struct poptOption * opt,
+		  const char * arg, void * data)
+{
+	GnomeProgram *program = data;
+
+	if (reason == POPT_CALLBACK_REASON_PRE)
+		g_dataset_set_data (con, "GnomeProgram", program);
+}
+
+
 /**
  * gnome_program_preinit:
  * @program: Application object
@@ -1188,16 +1269,14 @@ gnome_program_preinit (GnomeProgram *program,
 
     g_return_val_if_fail (program != NULL, NULL);
     g_return_val_if_fail (GNOME_IS_PROGRAM (program), NULL);
+    g_return_val_if_fail (argv != NULL, NULL);
 
     if (program->_priv->state != APP_CREATE_DONE)
 	return NULL;
 
-    /* On non-glibc systems, this is not set up for us.  */
-    if (!program_invocation_name) {
-	program_invocation_name = argv[0];
-	program_invocation_short_name = g_path_get_basename (program_invocation_name);
-    }
-
+    /* Store invocation name */
+    g_set_prgname (argv[0]);
+    
     /* 0. Misc setup */
     if (program->_priv->app_id)
 	g_free (program->_priv->app_id);
@@ -1207,7 +1286,15 @@ gnome_program_preinit (GnomeProgram *program,
 	g_free (program->_priv->app_version);
     program->_priv->app_version = g_strdup (app_version);
     program->_priv->argc = argc;
-    program->_priv->argv = argv;
+
+    /* Make a copy of argv, the thing is that while we know the
+     * original argv will live until the end of the program,
+     * there are those evil people out there that modify it.
+     * Also, this may be some other argv, think 'fake argv' here */
+    program->_priv->argv = g_new (char *, argc + 1);
+    for (i = 0; i < argc; i++)
+        program->_priv->argv[i] = g_strdup (argv[i]);
+    program->_priv->argv[argc] = NULL;
 
     if (!program_modules) {
 	program_modules = g_ptr_array_new();
@@ -1228,13 +1315,17 @@ gnome_program_preinit (GnomeProgram *program,
     /* 3. call the pre-init functions */
     for (i = 0; (a_module = g_ptr_array_index (program_modules, i)); i++) {
 	if (a_module->pre_args_parse) {
+#ifdef TIME_INIT
 	    GTimer *timer = g_timer_new ();
 	    g_timer_start (timer);
 	    g_print ("Running pre_args_parse for: %s ...", a_module->name); 
+#endif
 	    a_module->pre_args_parse (program, a_module);
+#ifdef TIME_INIT
 	    g_timer_stop (timer);
 	    g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 	    g_timer_destroy (timer);
+#endif
 	}
     }
 
@@ -1242,10 +1333,21 @@ gnome_program_preinit (GnomeProgram *program,
     {
 	struct poptOption includer = {NULL, '\0', POPT_ARG_INCLUDE_TABLE,
 				      NULL, 0, NULL, NULL};
+	struct poptOption callback =
+		{ NULL, '\0', POPT_ARG_CALLBACK | POPT_CBFLAG_PRE,
+		  &set_context_data, 0,
+		  /* GOD THIS IS EVIL!  But popt is so UTERLY FUCKED IT'S
+		   * NOT EVEN FUNNY.  For some reason it passes 'descrip'
+		   * as 'data' to the callback, and there is no other way
+		   * to pass data.  Fun, eh? */
+		  (const char *)program,
+		  NULL };
 
 	program->_priv->top_options_table = g_array_new
 	    (TRUE, TRUE, sizeof (struct poptOption));
 
+	g_array_append_val (program->_priv->top_options_table, callback);
+
 	/* Put the special popt table in first */
 	includer.arg = poptHelpOptions;
 	includer.descrip = N_("Help options");
@@ -1287,11 +1389,13 @@ gnome_program_preinit (GnomeProgram *program,
 
 /**
  * gnome_program_module_load:
- * @program: Application object
  * @mod_name: module name
  *
  * Loads a shared library that contains a
- * 'GnomeModuleInfo dynamic_module_info' structure.
+ * #GnomeModuleInfo dynamic_module_info structure.
+ *
+ * Returns: The #GnomeModuleInfo structure that was loaded, or %NULL if the
+ * module could not be loaded.
  */
 const GnomeModuleInfo *
 gnome_program_module_load (const char *mod_name)
@@ -1326,7 +1430,7 @@ gnome_program_module_load (const char *mod_name)
 
 /**
  * gnome_program_parse_args:
- * @app: Application object
+ * @program: Application object
  *
  * Description: Parses the command line arguments for the application
  */
@@ -1355,6 +1459,144 @@ gnome_program_parse_args (GnomeProgram *program)
     }
 }
 
+static char *
+find_accessibility_module (GnomeProgram *program, const char *libname)
+{
+	char *sub;
+	char *path;
+	char *fname;
+	char *retval;
+
+	fname = g_strconcat (libname, "." G_MODULE_SUFFIX, NULL);
+	sub = g_strconcat ("gtk-2.0/modules", G_DIR_SEPARATOR_S, fname, NULL);
+
+	path = gnome_program_locate_file (
+		program, GNOME_FILE_DOMAIN_LIBDIR, sub, TRUE, NULL);
+
+	g_free (sub);
+
+	if (path)
+		retval = path;
+	else
+		retval = gnome_program_locate_file (
+			program, GNOME_FILE_DOMAIN_LIBDIR,
+			fname, TRUE, NULL);
+
+	g_free (fname);
+
+	return retval;
+}
+
+static gboolean
+accessibility_invoke_module (GnomeProgram *program,
+			     const char   *libname,
+			     gboolean      init)
+{
+	GModule    *handle;
+	void      (*invoke_fn) (void);
+	const char *method;
+	gboolean    retval = FALSE;
+	char       *module_name;
+
+	if (init)
+		method = "gnome_accessibility_module_init";
+	else
+		method = "gnome_accessibility_module_shutdown";
+
+	module_name = find_accessibility_module (program, libname);
+
+	if (!module_name) {
+		g_warning ("Accessibility: failed to find module '%s' which "
+			   "is needed to make this application accessible",
+			   libname);
+
+	} else if (!(handle = g_module_open (module_name, G_MODULE_BIND_LAZY))) {
+		g_warning ("Accessibility: failed to load module '%s': '%s'",
+			   libname, g_module_error ());
+
+	} else if (!g_module_symbol (handle, method, (gpointer *)&invoke_fn)) {
+		g_warning ("Accessibility: error library '%s' does not include "
+			   "method '%s' required for accessibility support",
+			   libname, method);
+		g_module_close (handle);
+
+	} else {
+		retval = TRUE;
+		invoke_fn ();
+	}
+
+	g_free (module_name);
+
+	return retval;
+}
+
+static gboolean
+accessibility_invoke (GnomeProgram *program, gboolean init)
+{
+	GSList *l;
+	gboolean use_gui = FALSE;
+
+	if (!program->_priv->accessibility_modules)
+		return FALSE;
+
+	for (l = program->_priv->accessibility_modules; l; l = l->next) {
+		GnomeModuleInfo *module = l->data;
+		
+		if (!strcmp (module->name, "gtk")) {
+			accessibility_invoke_module (program, "libgail", init);
+			use_gui = TRUE;
+
+		} else if (!strcmp (module->name, "libgnomeui")) {
+			accessibility_invoke_module (program, "libgail-gnome", init);
+			use_gui = TRUE;
+		}
+	}
+
+	if (use_gui) {
+		accessibility_invoke_module (program, "libatk-bridge", init);
+	}
+
+	return TRUE;
+}
+
+
+static void
+accessibility_init (GnomeProgram *program)
+{
+	int i;
+	gboolean do_init;
+	const char *env_var;
+	GSList *list = NULL;
+
+	/* Seek the module list we need */
+	for (i = 0; i < program_modules->len; i++) {
+		GnomeModuleInfo *module = g_ptr_array_index (program_modules, i);
+		
+		if (!module)
+			continue;
+
+		if (!strcmp (module->name, "gtk"))
+			list = g_slist_prepend (list, module);
+
+		else if (!strcmp (module->name, "libgnomeui"))
+			list = g_slist_prepend (list, module);
+	}
+
+	program->_priv->accessibility_modules = list;
+	
+	do_init = FALSE;
+
+	if ((env_var = g_getenv (GNOME_ACCESSIBILITY_ENV)))
+		do_init = atoi (env_var);
+	else
+		do_init = gconf_client_get_bool (
+			gconf_client_get_default (),
+			GNOME_ACCESSIBILITY_KEY, NULL);
+
+	if (do_init)
+		accessibility_invoke (program, TRUE);
+}
+
 /**
  * gnome_program_postinit:
  * @program: Application object
@@ -1377,24 +1619,28 @@ gnome_program_postinit (GnomeProgram *program)
     /* Call post-parse functions */
     for (i = 0; (a_module = g_ptr_array_index(program_modules, i)); i++) {
 	if (a_module->post_args_parse) {
+#ifdef TIME_INIT
 	    GTimer *timer = g_timer_new ();
 	    g_timer_start (timer);
 	    g_print ("Running post_args_parse for: %s ...", a_module->name); 
+#endif
 	    a_module->post_args_parse (program, a_module);
+#ifdef TIME_INIT
 	    g_timer_stop (timer);
 	    g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 	    g_timer_destroy (timer);
+#endif
 	}
     }
 
+    /* Accessibility magic */
+    accessibility_init (program);
+
     /* Free up stuff we don't need to keep around for the lifetime of the app */
 
     /* Note! we cannot kill the program_modules array as that
      * may be needed later */
 
-    g_array_free (program->_priv->top_options_table, TRUE);
-    program->_priv->top_options_table = NULL;
-
     g_blow_chunks(); /* Try to compact memory usage a bit */
 
     program->_priv->state = APP_POSTINIT_DONE;
@@ -1408,10 +1654,12 @@ gnome_program_postinit (GnomeProgram *program)
  * @argc: The number of commmand line arguments contained in @argv.
  * @argv: A string array of command line arguments.
  * @first_property_name: The first item in a %NULL-terminated list of attribute
- * name/value.
+ * name and value pairs (so this will be an attribute name or %NULL).
  * @...: The continuation of a %NULL-terminated list of attribute name/value
  * pairs.
  *
+ * Initialises the current GNOME libraries for use by the application.
+ *
  * Returns: A #GnomeProgram instance representing the current application.
  */
 GnomeProgram *
@@ -1463,10 +1711,13 @@ gnome_program_initv (GType type,
     GnomeProgram *program;
     GnomeProgramClass *klass;
     int i;
+
+#ifdef TIME_INIT
     GTimer *global_timer = g_timer_new ();
 
     g_timer_start (global_timer);
     g_print ("Starting gnome_program_init:\n\n");
+#endif
 
     g_type_init ();
 
@@ -1474,6 +1725,7 @@ gnome_program_initv (GType type,
 
     if (!program_initialized) {
 	const char *ctmp;
+	const GnomeModuleInfo *libgnome_module;
 
 	program_module_list = g_ptr_array_new ();
 	program_modules = g_ptr_array_new ();
@@ -1481,12 +1733,16 @@ gnome_program_initv (GType type,
 	/* keep array NULL terminated */
 	g_ptr_array_add (program_modules, NULL);
 
-	/* Always register libgnome. */
-	gnome_program_module_register (libgnome_module_info_get ());
-
 	/* Register the requested modules. */
 	gnome_program_module_register (module_info);
 
+	/* 
+	 * make sure libgnome is always registered.
+	 */
+	libgnome_module = libgnome_module_info_get ();
+	if (!gnome_program_module_registered (libgnome_module))
+		gnome_program_module_register (libgnome_module);
+
 	/* Only load shlib modules and do all that other good
 	 * stuff when not setuid/setgid, for obvious reasons */
 	if (geteuid () == getuid () &&
@@ -1501,14 +1757,14 @@ gnome_program_initv (GType type,
 		    add_to_module_list (program_module_list, argv[i+1]);
 	    }
 
-	    if ((ctmp = g_getenv ("GNOME_MODULES")))
+	    ctmp = g_getenv ("GNOME_MODULES");
+	    if (ctmp != NULL)
 	        add_to_module_list (program_module_list, ctmp);
 	}
 
 	/*
 	 * Load all the modules.
 	 */
-
 	for (i = 0; i < program_module_list->len; i++) {
 	    gchar *modname = g_ptr_array_index (program_module_list, i);
 
@@ -1519,13 +1775,17 @@ gnome_program_initv (GType type,
 	    GnomeModuleInfo *a_module = g_ptr_array_index (program_modules, i);
 
 	    if (a_module && a_module->init_pass) {
+#ifdef TIME_INIT
 	        GTimer *timer = g_timer_new ();
 	        g_timer_start (timer);
 	        g_print ("Running init_pass for: %s ...", a_module->name); 
+#endif
 		a_module->init_pass (a_module);
+#ifdef TIME_INIT
 		g_timer_stop (timer);
 		g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 		g_timer_destroy (timer);
+#endif
 	    }
 	}
 
@@ -1536,13 +1796,17 @@ gnome_program_initv (GType type,
 	    GnomeModuleInfo *a_module = g_ptr_array_index (program_modules, i);
 
 	    if (a_module && a_module->class_init) {
+#ifdef TIME_INIT
 	        GTimer *timer = g_timer_new ();
 	        g_timer_start (timer);
 	        g_print ("Running class_init for: %s ...", a_module->name); 
+#endif
 		a_module->class_init (klass, a_module);
+#ifdef TIME_INIT
 		g_timer_stop (timer);
 		g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 		g_timer_destroy (timer);
+#endif
 	    }
 	}
     } else if ( ! gnome_program_module_registered (module_info)) {
@@ -1555,13 +1819,17 @@ gnome_program_initv (GType type,
 	    GnomeModuleInfo *a_module = g_ptr_array_index (program_modules, i);
 
 	    if (a_module && a_module->init_pass) {
+#ifdef TIME_INIT
 	        GTimer *timer = g_timer_new ();
 	        g_timer_start (timer);
 	        g_print ("Running init_pass for: %s ...", a_module->name); 
+#endif
 		a_module->init_pass (a_module);
+#ifdef TIME_INIT
 		g_timer_stop (timer);
 		g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 		g_timer_destroy (timer);
+#endif
 	    }
 	}
 
@@ -1573,13 +1841,17 @@ gnome_program_initv (GType type,
 	    GnomeModuleInfo *a_module = g_ptr_array_index (program_modules, i);
 
 	    if (a_module && a_module->class_init) {
+#ifdef TIME_INIT
 	        GTimer *timer = g_timer_new ();
 	        g_timer_start (timer);
 	        g_print ("Running class_init for: %s ...", a_module->name); 
+#endif
 		a_module->class_init (klass, a_module);
+#ifdef TIME_INIT
 		g_timer_stop (timer);
 		g_print ("done (%f seconds)\n", g_timer_elapsed (timer, NULL));
 		g_timer_destroy (timer);
+#endif
 	    }
 	}
     }
@@ -1598,9 +1870,11 @@ gnome_program_initv (GType type,
     gnome_program_parse_args (program);
     gnome_program_postinit (program);
 
+#ifdef TIME_INIT
     g_timer_stop (global_timer);
     g_print ("\nGlobal init done in: %f seconds\n\n", g_timer_elapsed (global_timer, NULL));
     g_timer_destroy (global_timer);
+#endif
 
     return program;
 }
